#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/ext2_fs.h>

/* */

#ifdef FIXBACK
#undef EXT2_SUPER_MAGIC 
#define EXT2_SUPER_MAGIC 0xEF54
#define EXT2_SUPER_MAGIC_BUFFALO  0xEF53
#else
#define EXT2_SUPER_MAGIC_BUFFALO  0xEF54
#endif

#define EXT2_SUPER_MAGIC_0000     0x0000
#define EXT2_SUPER_MAGIC_FFFF     0xFFFF
#define EXT2_MAX_BLOCKS_PER_GROUP  ((1 << 16) - 8)

#define CONVERT_BYTE_ORDER(type, var)  convert_byte_order(sizeof(type), (unsigned char *)&(var))

const off_t MASTER_SUPER_BLOCK_OFFSET = 1024;

int DEBUG_MODE = 0;

/* */

typedef struct ext2_super_block ext2_super_block;

typedef struct ext2fs {
  int fd;
  int need_byte_swap;
  ext2_super_block super;
  int block_groups_count;
} ext2fs;

/* */

void convert_byte_order(int size, unsigned char *p);

void ext2fs__initialize(ext2fs *fs);
int ext2fs__open(ext2fs *fs, const char *name);
int ext2fs__close(ext2fs *fs);
void ext2_super_block__convert_byte_order(ext2_super_block *s);
int ext2fs__read_super_block(ext2fs *fs, int no, ext2_super_block *super_block);
int ext2fs__write_super_block(ext2fs *fs, int no, ext2_super_block *super_block);

/* */

void convert_byte_order(int size, unsigned char *p)
{
  int i;
  unsigned char c1, c2;

  for (i = 0; i < size / 2; i++) {
    c1 = p[i];
    c2 = p[size - i - 1];
    p[i] = c2;
    p[size - i - 1] = c1;
  }
}

/* */

void ext2fs__initialize(ext2fs *fs)
{
  memset(&fs, 0, sizeof(fs));

  fs->fd = -1;
}

int ext2fs__open(ext2fs *fs, const char *name)
{
  __u32 blocks_count;
  __u32 first_data_block;
  __u32 blocks_per_group;
  size_t nread;

  fs->fd = open(name, O_RDWR);
  if (fs->fd == -1)
    return -1;

  memset(&fs->super, 0, sizeof(ext2_super_block));

  if (lseek(fs->fd, MASTER_SUPER_BLOCK_OFFSET, SEEK_SET) == (off_t)-1)
    return -1;
  
  nread = read(fs->fd, &fs->super, sizeof(ext2_super_block));
  if (nread != sizeof(ext2_super_block))
    return -1;

  {
    union {
      __u32 var_u32;
      __u8 var_u8[4];
    } u;
    __u16 magic;

    magic = fs->super.s_magic;
    fs->need_byte_swap = 0;

    u.var_u32 = 0x12345678;
    if (u.var_u8[0] == 0x12) {
      fs->need_byte_swap = !fs->need_byte_swap;
      CONVERT_BYTE_ORDER(__u16, magic);
    }

    if (magic == EXT2_SUPER_MAGIC || magic == EXT2_SUPER_MAGIC_BUFFALO) {
      ; // Nothing to do here.
    } else {
      CONVERT_BYTE_ORDER(__u16, magic);
      if (magic == EXT2_SUPER_MAGIC || magic == EXT2_SUPER_MAGIC_BUFFALO) {
        fs->need_byte_swap = !fs->need_byte_swap;
      } else {
        CONVERT_BYTE_ORDER(__u16, magic);
        fprintf(stderr, "Bad magic number: %04x\n", magic);
        return -1;
      }
    }
  }
  if (fs->need_byte_swap)
    ext2_super_block__convert_byte_order(&fs->super);

  blocks_per_group = fs->super.s_blocks_per_group;
  if (blocks_per_group == 0 || blocks_per_group > EXT2_MAX_BLOCKS_PER_GROUP)
    return -1;

  blocks_count = fs->super.s_blocks_count;
  first_data_block = fs->super.s_first_data_block;
  blocks_per_group = fs->super.s_blocks_per_group;

  fs->block_groups_count =
    (blocks_count - first_data_block + blocks_per_group - 1) / blocks_per_group;

  return 0;
}

int ext2fs__close(ext2fs *fs)
{
  if (close(fs->fd) == -1)
    return -1;

  return 0;
}

int test_root(int a, int b)
{
  if (a == 0)
    return 1;

  while (1) {
    if (a == 1)
      return 1;
    if (a % b != 0)
      return 0;
    a = a / b;
  }

  // Never reach here
  abort();

  return 0;
}

int ext2fs__has_block_group_super_block(ext2fs *fs, int no)
{
  if (!(fs->super.s_feature_ro_compat & EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
    return 1;

  if (test_root(no, 3) || test_root(no, 5) || test_root(no, 7))
    return 1;

  return 0;
}

int ext2fs__read_super_block(ext2fs *fs, int no, ext2_super_block *s)
{
  __u32 block_group_loc;
  off_t super_block_offset;
  off_t lseek_result;
  size_t nread;
  int has_super_block;

  if (no < 0)
    return 1;

  has_super_block = ext2fs__has_block_group_super_block(fs, no);
  if (!has_super_block)
    return 1;

  block_group_loc = fs->super.s_first_data_block + fs->super.s_blocks_per_group * no;

  super_block_offset = (off_t)EXT2_BLOCK_SIZE(&fs->super) * block_group_loc;
  if (no == 0)
    super_block_offset = MASTER_SUPER_BLOCK_OFFSET;

  lseek_result = lseek(fs->fd, super_block_offset, SEEK_SET);
  if (lseek_result != super_block_offset)
    return -1;

  nread = read(fs->fd, s, sizeof(ext2_super_block));
  if (nread != sizeof(ext2_super_block))
    return -1;

  if (fs->need_byte_swap)
    ext2_super_block__convert_byte_order(s);

  if (DEBUG_MODE) {
    printf("\n");
    printf("(debug) super_block_offset = 0x%llx\n", (long long)super_block_offset);
    printf("(debug) uuid = %02x%02x%02x%02x-%02x%02x-%02x%02x-%02x%02x-%02x%02x%02x%02x%02x%02x\n",
      s->s_uuid[ 0], s->s_uuid[ 1], s->s_uuid[ 2], s->s_uuid[ 3],
      s->s_uuid[ 4], s->s_uuid[ 5], s->s_uuid[ 6], s->s_uuid[ 7],
      s->s_uuid[ 8], s->s_uuid[ 9], s->s_uuid[10], s->s_uuid[11],
      s->s_uuid[12], s->s_uuid[13], s->s_uuid[14], s->s_uuid[15]
    );
  }

  return 0;
}

int ext2fs__write_super_block(ext2fs *fs, int no, ext2_super_block *s)
{
  __u32 block_group_loc;
  off_t super_block_offset;
  off_t lseek_result;
  size_t nwrote;
  int has_super_block;

  if (no < 0)
    return 1;

  has_super_block = ext2fs__has_block_group_super_block(fs, no);
  if (!has_super_block)
    return 1;

  block_group_loc = fs->super.s_first_data_block + fs->super.s_blocks_per_group * no;

  super_block_offset = (off_t)EXT2_BLOCK_SIZE(&fs->super) * block_group_loc;
  if (no == 0)
    super_block_offset = MASTER_SUPER_BLOCK_OFFSET;

  lseek_result = lseek(fs->fd, super_block_offset, SEEK_SET);
  if (lseek_result != super_block_offset)
    return -1;

  if (fs->need_byte_swap)
    ext2_super_block__convert_byte_order(s);

  nwrote = write(fs->fd, s, sizeof(ext2_super_block));
  if (nwrote != sizeof(ext2_super_block))
    return -1;

  return 0;
}

void ext2_super_block__convert_byte_order(ext2_super_block *s)
{
  CONVERT_BYTE_ORDER(__u32, s->s_inodes_count);
  CONVERT_BYTE_ORDER(__u32, s->s_blocks_count);
  CONVERT_BYTE_ORDER(__u32, s->s_r_blocks_count);
  CONVERT_BYTE_ORDER(__u32, s->s_free_blocks_count);
  CONVERT_BYTE_ORDER(__u32, s->s_free_inodes_count);
  CONVERT_BYTE_ORDER(__u32, s->s_first_data_block);
  CONVERT_BYTE_ORDER(__u32, s->s_log_block_size);
  CONVERT_BYTE_ORDER(__s32, s->s_log_frag_size);
  CONVERT_BYTE_ORDER(__u32, s->s_blocks_per_group);
  CONVERT_BYTE_ORDER(__u32, s->s_frags_per_group);
  CONVERT_BYTE_ORDER(__u32, s->s_inodes_per_group);
  CONVERT_BYTE_ORDER(__u32, s->s_mtime);
  CONVERT_BYTE_ORDER(__u32, s->s_wtime);
  CONVERT_BYTE_ORDER(__u16, s->s_mnt_count);
  CONVERT_BYTE_ORDER(__s16, s->s_max_mnt_count);
  CONVERT_BYTE_ORDER(__u16, s->s_magic);
  CONVERT_BYTE_ORDER(__u16, s->s_state);
  CONVERT_BYTE_ORDER(__u16, s->s_errors);
  CONVERT_BYTE_ORDER(__u16, s->s_minor_rev_level);
  CONVERT_BYTE_ORDER(__u32, s->s_lastcheck);
  CONVERT_BYTE_ORDER(__u32, s->s_checkinterval);
  CONVERT_BYTE_ORDER(__u32, s->s_creator_os);
  CONVERT_BYTE_ORDER(__u32, s->s_rev_level);
  CONVERT_BYTE_ORDER(__u16, s->s_def_resuid);
  CONVERT_BYTE_ORDER(__u16, s->s_def_resgid);
  CONVERT_BYTE_ORDER(__u32, s->s_first_ino);
  CONVERT_BYTE_ORDER(__u16, s->s_inode_size);
  CONVERT_BYTE_ORDER(__u16, s->s_block_group_nr);
  CONVERT_BYTE_ORDER(__u32, s->s_feature_compat);
  CONVERT_BYTE_ORDER(__u32, s->s_feature_incompat);
  CONVERT_BYTE_ORDER(__u32, s->s_feature_ro_compat);
  CONVERT_BYTE_ORDER(__u32, s->s_algorithm_usage_bitmap);
}

/* */

#define PROGRAM_NAME  "fix_ex2_magic"
#define PROGRAM_VERSION  "1.1"

static volatile void print_usage_and_die()
{
  printf("%s Version %s - show/fix ext2 magic number\n", PROGRAM_NAME, PROGRAM_VERSION);
  printf("Usage: %s [--debug|--fix] <device file|disk image file>\n", PROGRAM_NAME);

  exit(1);
}

int main(int argc, char *argv[])
{
  ext2fs fs;
  const char *device_file_name;
  int do_fix_magic_number;
  int no;

  do_fix_magic_number = 0;
  if (argc == 1 || argc > 3)
    print_usage_and_die();
  if (argc == 2 && argv[1][0] != '-') {
    device_file_name = argv[1];
  } else {
    if (strcmp(argv[1], "--fix") == 0)
      do_fix_magic_number = 1;
    else if (strcmp(argv[1], "--debug") == 0)
      DEBUG_MODE = 1;
    else
      print_usage_and_die();
    device_file_name = argv[2];
  }

  if (ext2fs__open(&fs, device_file_name) == -1) {
    printf("Can't open %s\n", device_file_name);
    return 1;
  }

  for(no = 0; no < fs.block_groups_count; no++) {
    ext2_super_block s;
    int result;

    result = ext2fs__read_super_block(&fs, no, &s);
    if (result == -1) {
      printf("Can't read super block\n");
      printf("Block group %d\n", no);
      return 1;
    }
    if (result == 0) {
      printf("Block group %3d: magic number = 0x%04x", no, s.s_magic);

      if (!DEBUG_MODE) {
        if (s.s_magic != EXT2_SUPER_MAGIC && s.s_magic != EXT2_SUPER_MAGIC_BUFFALO) {
          printf("\n");
          printf("Unknown magic number\n");
          return 1;
        }
      }

      if (do_fix_magic_number && s.s_magic == EXT2_SUPER_MAGIC_BUFFALO) {
        s.s_magic = EXT2_SUPER_MAGIC;
        printf(" -> 0x%04x (fixed)", s.s_magic);

        result = ext2fs__write_super_block(&fs, no, &s);
        if (result != 0) {
          printf("\n");
          printf("Can't write super block\n");
          return 1;
        }
      }

      printf("\n");
    }
  }

  if (ext2fs__close(&fs) == -1) {
    printf("Can't close %s\n", device_file_name);
    return 1;
  }

  return 0;
}
